/* GNU POP3 - a small, fast, and efficient POP3 daemon
   Copyright (C) 1999 Jakob 'sparky' Kaivo <jkaivo@nodomainname.net>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include "gnu-pop3d.h"

#ifdef USE_PAM
#define COPY_STRING(s) (s) ? strdup(s) : NULL

static char *_pwd;
static char *_user;
static int _perr = 0;

static int
PAM_gnupop3d_conv (int num_msg, const struct pam_message **msg,
				   struct pam_response **resp, void *appdata_ptr)
{
  int replies = 0;
  struct pam_response *reply = NULL;

  reply = (struct pam_response *) malloc (sizeof (struct pam_response) * num_msg);
  if (!reply)
	return PAM_CONV_ERR;

  for (replies = 0; replies < num_msg; replies++)
	{
	  switch (msg[replies]->msg_style)
		{
		case PAM_PROMPT_ECHO_ON:
		  reply[replies].resp_retcode = PAM_SUCCESS;
		  reply[replies].resp = COPY_STRING (_user);
		  /* PAM frees resp */
		  break;
		case PAM_PROMPT_ECHO_OFF:
		  reply[replies].resp_retcode = PAM_SUCCESS;
		  reply[replies].resp = COPY_STRING (_pwd);
		  /* PAM frees resp */
		  break;
		case PAM_TEXT_INFO:
		case PAM_ERROR_MSG:
		  reply[replies].resp_retcode = PAM_SUCCESS;
		  reply[replies].resp = NULL;
		  break;
		default:
		  free (reply);
		  _perr = 1;
		  return PAM_CONV_ERR;
		}
	}
  *resp = reply;
  return PAM_SUCCESS;
}

static struct pam_conv PAM_conversation = { &PAM_gnupop3d_conv, NULL };

#endif /* USE_PAM */

/* Basic user authentication. This also takes the PASS command and verifies
   the user name and password. Calls setuid() upon successful verification,
   otherwise it will (likely) return ERR_BAD_LOGIN */

int
pop3_user (const char *arg)
{
#ifdef USE_VIRTUAL
  char *domainname;
#endif
  char *buf, pass[POP_MAXCMDLEN], *tmp, *cmd;

#ifdef USE_PAM
  pam_handle_t *pamh;
  int pamerror;

#endif /* !USE_PAM */
  struct passwd *pw = NULL;

  if (state != AUTHORIZATION)
	return ERR_WRONG_STATE;

  if ((strlen (arg) == 0) || (strchr (arg, ' ') != NULL))
	return ERR_BAD_ARGS;

  fprintf (ofile, "+OK\r\n");
  fflush (ofile);

  if ((username = strdup (arg)) == NULL)
	pop3_abquit (ERR_NO_MEM);

#ifdef USE_VIRTUAL
  domainname = pop3_splitname (username, useroptions);
  virtualdomain = domainname;
#ifdef DEFAULT_DOMAIN
  if (virtualdomain)
	if (strcasecmp (DEFAULT_DOMAIN, virtualdomain) == 0)
	  {
		/* fallback to normal */
		free (virtualdomain);
		virtualdomain = NULL;
		domainname = NULL;
	  }
#endif /* DEFAULT_DOMAIN */
#endif /* USE_VIRTUAL */

  buf = pop3_readline (ifile);
  cmd = pop3_cmd (buf);
  tmp = pop3_args (buf);
  free (buf);

  if (strlen (tmp) > POP_MAXCMDLEN)
	{
	  free (tmp);
	  return ERR_TOO_LONG;
	}
  else
	{
	  strncpy (pass, tmp, POP_MAXCMDLEN);
	  free (tmp);
	}

  if (strlen (cmd) > 4)
	{
	  free (cmd);
	  return ERR_BAD_CMD;
	}
  if ((strcasecmp (cmd, "PASS") == 0))
	{
	  free (cmd);

#ifdef USE_VIRTUAL
/* uses regular crypt and not PAM for virtual domains authentication */

	  if (virtualdomain)
		{
		  pw = getvirtualpwnam (username, virtualdomain);

		  if (pw == NULL)
			{
			  syslog (LOG_INFO, "Non-existent user: %s (%s)", arg, virtualdomain);
			  return ERR_BAD_LOGIN;
			}
		  if (pw->pw_uid < 1)
			{
			  pop3_freepw (pw);
			  syslog (LOG_INFO, "User id less than 1 for %s (%s)", arg, virtualdomain);
			  return ERR_BAD_LOGIN;
			}
		  if (debug & 0x80)
			syslog (LOG_INFO, "shell=%s pw_p=%s pass=%s crypt(%s)", pw->pw_shell, pw->pw_passwd,
					pass, crypt (pass, pw->pw_passwd));
		  if ((strcmp (pw->pw_shell, "encrypted") && strcmp (pw->pw_passwd, pass))
			  || (!strcmp (pw->pw_shell, "encrypted")
				  && strcmp (pw->pw_passwd, crypt (pass, pw->pw_passwd))))
			{
			  pop3_freepw (pw);
			  syslog (LOG_INFO, "Incorrect encrypted password for %s (%s)", arg, virtualdomain);
			  return ERR_BAD_LOGIN;
			}
		}
	  else
		{						/* no domain part so revert to normal */
#endif /* USE_VIRTUAL */
		  pw = getpwnam (username);

#ifndef USE_PAM
		  if (pw == NULL)
			{
			  syslog (LOG_INFO, "Non-existent user: %s", arg);
			  return ERR_BAD_LOGIN;
			}
		  if (pw->pw_uid < 1)
			{
#ifdef USE_VIRTUAL
			  if (virtualdomain)
				pop3_freepw (pw);
#endif
			  syslog (LOG_INFO, "User id less than 1 for %s", arg);
			  return ERR_BAD_LOGIN;
			}
		  if (strcmp (pw->pw_passwd, crypt (pass, pw->pw_passwd)))
			{
#ifdef HAVE_SHADOW_H
			  struct spwd *spw;

			  spw = getspnam (arg);
			  if (spw == NULL)
				{
				  syslog (LOG_INFO, "Could not access shadow password for %s", arg);
				  return ERR_BAD_LOGIN;
				}
			  if (strcmp (spw->sp_pwdp, crypt (pass, spw->sp_pwdp)))
				{
#endif /* HAVE_SHADOW_H */
				  syslog (LOG_INFO, "Incorrect encrypted password for %s", arg);
				  return ERR_BAD_LOGIN;
				}
			}
#else /* USE_PAM */
		  _user = (char *) username;
		  _pwd = pass;
		  if (debug & 0x40)
			syslog (LOG_INFO, "pam: us=%s", _user);
		  /* libpam doesn't log to LOG_MAIL */
		  closelog ();
		  pamerror = pam_start ("gnu-pop3d", username, &PAM_conversation, &pamh);
		  if (!_perr && (pamerror == PAM_SUCCESS))
			{
			  pamerror = pam_authenticate (pamh, 0);
			  if (!_perr && (pamerror == PAM_SUCCESS))
				{
				  pamerror = pam_acct_mgmt (pamh, 0);
				  if (!_perr && (pamerror == PAM_SUCCESS))
					pamerror = pam_setcred (pamh, PAM_ESTABLISH_CRED);
				}
			}
		  pam_end (pamh, PAM_SUCCESS);
		  openlog ("gnu-pop3d", LOG_PID, LOG_MAIL);
		  if (_perr || (pamerror != PAM_SUCCESS))
			{
			  syslog (LOG_INFO, "pam: %s", pam_strerror (pamh, pamerror));
			  return ERR_BAD_LOGIN;
			}
#endif /* USE_PAM */

#ifdef USE_VIRTUAL
		}						/* has no domain part */
#endif /* USE_VIRTUAL */
	  return pop3_finishlogin (pw);
	}
  else if (strcasecmp (cmd, "QUIT") == 0)
	{
	  free (cmd);
	  return pop3_quit (pass);
	}
#ifdef USE_VIRTUAL
  if (virtualdomain)
	pop3_freepw (pw);
#endif /* USE_VIRTUAL */
  syslog (LOG_INFO, "Incorrect command after USER: %s", cmd);
  free (cmd);
  return ERR_BAD_LOGIN;
}

/* The password been validated OK.
   Set user.group and get mailbox */
int
pop3_finishlogin (struct passwd *pw)
{
  struct sockaddr_in peername;
  socklen_t namelen = sizeof (peername);
  int rc;
  char *pos;

  if (debug & 0x20)
	syslog (LOG_INFO, "user's uid %d gid %d", (pw == NULL) ? -1 : pw->pw_uid,
			(pw == NULL) ? -1 : pw->pw_gid);

  if ((pw == NULL) || pw->pw_uid < 1 || setuid (pw->pw_uid) == -1)
	{
	  pop3_unlock ();
#ifdef USE_VIRTUAL
	  if (virtualdomain)
		pop3_freepw (pw);
#endif
	  syslog (LOG_ERR, "uid or gid not set: uid %d gid %d",
			  (pw == NULL) ? -1 : pw->pw_uid, (pw == NULL) ? -1 : pw->pw_gid);
	  syslog (LOG_INFO, "current uid %d, gid %d", getuid (), getgid ());
	  return ERR_BAD_LOGIN;
	}

  if (debug & 0x40)
	syslog (LOG_INFO, "current uid %d, gid %d", getuid (), getgid ());

  mbox = NULL;

#ifdef MAILSPOOLHOME
  mailbox = malloc ((strlen (pw->pw_dir) + strlen (MAILSPOOLHOME) + 1) * sizeof (char));
  if (mailbox == NULL)
	pop3_abquit (ERR_NO_MEM);
  if ((pos = strstr (pw->pw_dir, "/./")) != NULL)
	*pos = '\0';
  strcpy (mailbox, pw->pw_dir);
  strcat (mailbox, MAILSPOOLHOME);
  mbox = fopen (mailbox, "r+");
  if (debug & 0x40)
	syslog (LOG_INFO, "mailbox: %s %s errno=%i", mailbox, (mbox == NULL) ? "NOT FOUND" : "FOUND",
			errno);
  if (mbox == NULL)
	{
	  free (mailbox);
	  mailbox = NULL;
	}
#endif /* MAILSPOOLHOME */

  if (mbox == NULL)
	{
#ifdef USE_VIRTUAL
	  if (virtualdomain)
		{
		  mailbox =
			malloc ((strlen (VIRTUAL_MAILPATH) + 1 + strlen (virtualdomain) +
					 1 + strlen (username) + 1) * sizeof (char));
		  if (mailbox == NULL)
			pop3_abquit (ERR_NO_MEM);
		  strcpy (mailbox, VIRTUAL_MAILPATH);
		  strcat (mailbox, "/");
		  strcat (mailbox, virtualdomain);
		}
	  else
		{
#endif
		  mailbox = malloc ((strlen (_PATH_MAILDIR) + 1 + strlen (username) + 1) * sizeof (char));
		  if (mailbox == NULL)
			pop3_abquit (ERR_NO_MEM);
		  strcpy (mailbox, _PATH_MAILDIR);
#ifdef USE_VIRTUAL
		}
#endif
	  strcat (mailbox, "/");
	  strcat (mailbox, username);
	  mbox = fopen (mailbox, "r+");
	  if (debug & 0x40)
		syslog (LOG_INFO, "mailbox: %s %s errno=%i", mailbox,
				(mbox == NULL) ? "NOT FOUND" : "FOUND", errno);
	}

#ifdef USE_VIRTUAL
  /* maybe the postmaster with the real user in the passwd gecos field ? */
  if ((mbox == NULL) && virtualdomain)
	{
	  mailbox = malloc ((strlen (_PATH_MAILDIR) + 1 + strlen (pw->pw_gecos) + 1) * sizeof (char));
	  if (mailbox == NULL)
		pop3_abquit (ERR_NO_MEM);
	  strcpy (mailbox, _PATH_MAILDIR);
	  strcat (mailbox, "/");
	  strcat (mailbox, pw->pw_gecos);
	  mbox = fopen (mailbox, "r+");
	  if (debug & 0x40)
		syslog (LOG_INFO, "mailbox: %s %s errno=%i", mailbox,
				(mbox == NULL) ? "NOT FOUND" : "FOUND", errno);
	}
#endif

#ifdef USE_VIRTUAL
  /* maybe simply a mailbox and not a dir ? */
  if ((mbox == NULL) && virtualdomain)
	{
	  pos = pw->pw_dir + strlen (pw->pw_dir) - 1;
	  if (pos != strrchr (pw->pw_dir, '/'))
		{						/* probably not a dir */
		  mailbox = malloc ((strlen (pw->pw_dir) + 1) * sizeof (char));
		  if (mailbox == NULL)
			pop3_abquit (ERR_NO_MEM);
		  strcpy (mailbox, pw->pw_dir);
		  mbox = fopen (mailbox, "r+");
		  if (debug & 0x40)
			syslog (LOG_INFO, "mailbox: %s %s errno=%i", mailbox,
					(mbox == NULL) ? "NOT FOUND" : "FOUND", errno);
		}
	}
#endif
  if (mbox == NULL)
	{
	  syslog (LOG_ERR, "Unable to use %s mailbox; using /dev/null for testing", mailbox);
	  free (mailbox);
	  mailbox = malloc (strlen ("/dev/null") + 1);
	  if (mailbox == NULL)
		{
		  free (username);
		  username = NULL;
		  pop3_abquit (ERR_NO_MEM);
		}
	  strcpy (mailbox, "/dev/null");
	  mbox = fopen (mailbox, "r+");
	  if (debug & 0x40)
		syslog (LOG_INFO, "mailbox: %s %s errno=%i", mailbox,
				(mbox == NULL) ? "NOT FOUND" : "FOUND", errno);
	}
  else if ((rc = pop3_lock ()) != OK)
	{
	  pop3_unlock ();
	  free (mailbox);
	  mailbox = NULL;
#ifdef USE_VIRTUAL
	  if (virtualdomain)
		pop3_freepw (pw);
#endif
	  state = AUTHORIZATION;
	  return rc;
	}

  pop3_recover_oldmailbox (mailbox, mbox);

  state = TRANSACTION;
  messages = NULL;
  pop3_getsizes ();

#ifdef USE_VIRTUAL
  if (virtualdomain)
	{
	  char *tmp;
	  tmp = calloc (strlen (username) + 1 + strlen (virtualdomain) + 1, sizeof (char));
	  if (tmp == NULL)
		{
		  free (username);
		  username = NULL;
		  free (mailbox);
		  mailbox = NULL;
		  pop3_abquit (ERR_NO_MEM);
		}
	  tmp = strcat (tmp, username);
	  tmp = strcat (tmp, "@");
	  free (username);
	  tmp = strcat (tmp, virtualdomain);
	  username = tmp;
	}
#endif /* USE_VIRTUAL */

  fprintf (ofile, "+OK opened mailbox for %s\r\n", username);
  getpeername (ifile, (struct sockaddr *) &peername, &namelen);
  syslog (LOG_INFO, "User %s logged in with mailbox %s from %s", username,
		  mailbox, inet_ntoa (peername.sin_addr));

  cursor = 0;
#ifdef USE_VIRTUAL
  if (virtualdomain)
	pop3_freepw (pw);
#endif /* USE_VIRTUAL */
  return OK;
}
